/* Copyright 2020 Purdea Andrei
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include QMK_KEYBOARD_H

// Defines names for use in layer keycodes and the keymap
enum layer_names {
    _BASE,
    _FN,
    _FN2
};

// Defines the keycodes used by our macros in process_record_user
enum custom_keycodes {
    QMKBEST = SAFE_RANGE,
    QMKURL,
    QMKADM,
    QMKSTD,
    QMKROL,
    QMKHOM,
    QMKOA
};

typedef struct {
    bool is_press_action;
    uint8_t state;
} tap;

// Define a type for as many tap dance states I need
enum {
    SINGLE_TAP = 1,
    SINGLE_HOLD,
    DOUBLE_TAP,
    DOUBLE_HOLD,
    DOUBLE_SINGLE_TAP, // Send two single taps
    TRIPLE_TAP,
    TRIPLE_HOLD
};

enum {
            TD_DOT_COMM,
            INS_LAYR,
            TD_ESC,
	    TD_FNSH
};

uint8_t cur_dance(qk_tap_dance_state_t *state);

// For the x tap dance. Put it here so it can be used in any keymap
void x_finished(qk_tap_dance_state_t *state, void *user_data);
void x_reset(qk_tap_dance_state_t *state, void *user_data);


const uint16_t PROGMEM keymaps[][MATRIX_ROWS][MATRIX_COLS] = {
    /* Base */
    [_BASE] = LAYOUT_all(
        TD(TD_ESC), KC_1,    KC_2,    KC_3,    KC_4,    KC_5,     KC_6,    KC_7,    KC_8,    KC_9,    KC_0,    KC_MINS,   KC_EQL,  KC_NO, KC_BSPC,
        KC_TAB,     KC_Q,    KC_W,    KC_E,    KC_R,    KC_T,     KC_Y,    KC_U,    KC_I,    KC_O,    KC_P,    KC_LBRC,   KC_RBRC, KC_BSLS,
        KC_LCTL,    KC_A,    KC_S,    KC_D,    KC_F,    KC_G,     KC_H,    KC_J,    KC_K,    KC_L,    KC_SCLN, KC_QUOT,   KC_NO,   KC_ENT,
        KC_LSFT,    KC_NO,   KC_Z,    KC_X,    KC_C,    KC_V,    KC_B,     KC_N,    KC_M,    KC_COMM, KC_DOT,  KC_SLSH, RSFT_T(KC_UP), TD(TD_FNSH),
        KC_LALT, MO(_FN2), KC_LGUI,                                KC_SPC,                   KC_RGUI, KC_LEFT, KC_DOWN,   KC_RGHT
    ),
    [_FN] = LAYOUT_all(
        KC_GRV,     KC_F1,   KC_F2,   KC_F3,   KC_F4,   KC_F5,    KC_F6,   KC_F7,   KC_F8,   KC_F9,   KC_F10,   KC_F11,   KC_F12,  KC_NO, KC_DEL,
        _______,    QMKBEST, QMKOA, _______, QMKROL,  _______,  _______, _______, KC_PSCR, KC_SLCK, KC_PAUS,  HPT_DWLI, HPT_DWLD, HPT_TOG,
        _______,    QMKADM,  QMKSTD,  _______, _______,  _______,  QMKHOM,  _______, _______, _______, _______,  _______, KC_NO,   _______,
        _______,    KC_NO,   _______, _______, _______, _______, _______,  _______, KC_MUTE, KC_VOLD, KC_VOLU, _______,  KC_PGUP, KC_TRNS,
        KC_LALT, KC_TRNS, KC_LGUI,                                _______,                   KC_RGUI, KC_HOME,  KC_PGDN, KC_END
    ),
    [_FN2] = LAYOUT_all(
        KC_GRV,     KC_F1,   KC_F2,   KC_F3,   KC_F4,   KC_F5,    KC_F6,   KC_F7,   KC_F8,   KC_F9,   KC_F10,   KC_F11,   KC_F12,  KC_NO, KC_DEL,
        _______,    QMKBEST, _______, _______, QMKROL,  _______,  _______, _______, KC_PSCR, KC_SLCK, KC_PAUS,  _______, _______, _______,
        _______,    QMKADM,  QMKSTD,  _______, _______,  _______,  QMKHOM,  _______, _______, _______, _______,  _______, KC_NO,   _______,
        _______,    KC_NO,   _______, _______, _______, _______, _______,  _______, KC_MUTE, KC_VOLD, KC_VOLU, _______,  KC_PGUP,  KC_TRNS,
        KC_LALT, KC_TRNS, KC_LGUI,                                _______,                   KC_RGUI, KC_HOME,  KC_PGDN,  KC_END
    ),
};

bool process_record_user(uint16_t keycode, keyrecord_t *record) {
    switch (keycode) {
        case QMKBEST:
            if (record->event.pressed) {
                // when keycode QMKBEST is pressed
                SEND_STRING("QBR");
            } else {
                // when keycode QMKBEST is released
            }
            break;
        case QMKROL:
            if (record->event.pressed) {
                // when keycode QMKROL is pressed
                SEND_STRING("Rolling");
            } else {
                // when keycode QMROL is released
            }
            break;
         case QMKADM:
            if (record->event.pressed) {
                // when keycode QMKADM is pressed
                SEND_STRING("Administrator");
            } else {
                // when keycode QMKADM is released
            }
            break;
         case QMKSTD:
            if (record->event.pressed) {
                // when keycode QMKSTD is pressed
                SEND_STRING("Standard");
            } else {
                // when keycode QMKSTD is released
            }
            break;
         case QMKHOM:
            if (record->event.pressed) {
                // when keycode QMKHOM is pressed
                SEND_STRING("Home");
            } else {
                // when keycode QMKHOM is released
            }
            break;
	    case QMKOA:
            if (record->event.pressed) {
                // when keycode QMKOA is pressed
                SEND_STRING("Operational Access");
            } else {
                // when keycode QMKOA is released
            }
            break;
    }
    return true;
}

uint8_t cur_dance(qk_tap_dance_state_t *state) {
    if (state->count == 1) {
        if (state->interrupted || !state->pressed) return SINGLE_TAP;
        // Key has not been interrupted, but the key is still held. Means you want to send a 'HOLD'.
        else return SINGLE_HOLD;
    } else if (state->count == 2) {
        // DOUBLE_SINGLE_TAP is to distinguish between typing "pepper", and actually wanting a double tap
        // action when hitting 'pp'. Suggested use case for this return value is when you want to send two
        // keystrokes of the key, and not the 'double tap' action/macro.
        if (state->interrupted) return DOUBLE_SINGLE_TAP;
        else if (state->pressed) return DOUBLE_HOLD;
        else return DOUBLE_TAP;
    }

    // Assumes no one is trying to type the same letter three times (at least not quickly).
    // If your tap dance key is 'KC_W', and you want to type "www." quickly - then you will need to add
    // an exception here to return a 'TRIPLE_SINGLE_TAP', and define that enum just like 'DOUBLE_SINGLE_TAP'
    if (state->count == 3) {
        if (state->interrupted || !state->pressed) return TRIPLE_TAP;
        else return TRIPLE_HOLD;
    } else return 8; // Magic number. At some point this method will expand to work for more presses
}

// Create an instance of 'tap' for the 'x' tap dance.
static tap xtap_state = {
    .is_press_action = true,
    .state = 0
};

void x_finished(qk_tap_dance_state_t *state, void *user_data) {
    xtap_state.state = cur_dance(state);
    switch (xtap_state.state) {
        case SINGLE_TAP: register_code(KC_NO); break;
        case SINGLE_HOLD: layer_on(_FN); break;
        case DOUBLE_TAP: register_code(KC_NO); break;
        case DOUBLE_HOLD: register_code(KC_RSFT); break;
        // Last case is for fast typing. Assuming your key is `f`:
        // For example, when typing the word `buffer`, and you want to make sure that you send `ff` and not `Esc`.
        // In order to type `ff` when typing fast, the next character will have to be hit within the `TAPPING_TERM`, which by default is 200ms.
        // case DOUBLE_SINGLE_TAP: tap_code(KC_NO); register_code(KC_NO);
    }
}

void x_reset(qk_tap_dance_state_t *state, void *user_data) {
    switch (xtap_state.state) {
        case SINGLE_TAP: unregister_code(KC_NO); break;
        case SINGLE_HOLD: layer_off(_FN); break;
        case DOUBLE_TAP: unregister_code(KC_NO); break;
        case DOUBLE_HOLD: unregister_code(KC_RSFT);
        // case DOUBLE_SINGLE_TAP: unregister_code(KC_NO);
    }
    xtap_state.state = 0;
}

// Associate our tap dance key with its functionality
qk_tap_dance_action_t tap_dance_actions[] = {
    [TD_ESC] = ACTION_TAP_DANCE_DOUBLE(KC_ESC, KC_GRV),
    [TD_FNSH] = ACTION_TAP_DANCE_FN_ADVANCED(NULL, x_finished, x_reset),
};
